@@ -34,6 +34,10 @@
 #include <stdint.h>
 #include <common.h>
 
+#ifdef _WIN32
+# define sleep(x) Sleep(x*1000)
+#endif
+
 static
 char *get_single_token_url(common_info_st * info);
 
@@ -41,7 +45,7 @@ char *get_single_token_url(common_info_st * info);
 	if (url == NULL) { \
 		url = get_single_token_url(info); \
 		if (url == NULL) { \
-			fprintf(stderr, "warning: no token URL was provided for this operation; the available tokens are:\n"); \
+			fprintf(stderr, "warning: no token URL was provided for this operation; the available tokens are:\n\n"); \
 			pkcs11_token_list(out, det, info, 1); \
 			exit(1); \
 		} \
@@ -92,22 +96,29 @@ pkcs11_list(FILE * outfile, const char *url, int type, unsigned int flags,
 	    unsigned int detailed, common_info_st * info)
 {
 	gnutls_pkcs11_obj_t *crt_list;
-	unsigned int crt_list_size = 0, i;
+	unsigned int crt_list_size = 0, i, j;
 	int ret, otype;
 	char *output, *str;
-	int attrs;
+	int attrs, print_exts = 0;
+	gnutls_x509_ext_st *exts;
+	unsigned exts_size;
 	unsigned int obj_flags = flags;
 
 	pkcs11_common(info);
 
 	FIX(url, outfile, detailed, info);
 
+	gnutls_pkcs11_token_get_flags(url, &flags);
+	if (flags & GNUTLS_PKCS11_TOKEN_TRUSTED)
+		print_exts = 1;
+
 	if (type == PKCS11_TYPE_TRUSTED) {
 		attrs = GNUTLS_PKCS11_OBJ_ATTR_CRT_TRUSTED;
 	} else if (type == PKCS11_TYPE_PK) {
 		attrs = GNUTLS_PKCS11_OBJ_ATTR_CRT_WITH_PRIVKEY;
 	} else if (type == PKCS11_TYPE_CRT_ALL) {
 		attrs = GNUTLS_PKCS11_OBJ_ATTR_CRT_ALL;
+		if (print_exts != 0) print_exts++;
 	} else if (type == PKCS11_TYPE_PRIVKEY) {
 		attrs = GNUTLS_PKCS11_OBJ_ATTR_PRIVKEY;
 	} else if (type == PKCS11_TYPE_INFO) {
@@ -129,7 +140,7 @@ pkcs11_list(FILE * outfile, const char *url, int type, unsigned int flags,
 
 	if (crt_list_size == 0) {
 		fprintf(stderr, "No matching objects found\n");
-		exit(0);
+		exit(2);
 	}
 
 	for (i = 0; i < crt_list_size; i++) {
@@ -187,12 +198,141 @@ pkcs11_list(FILE * outfile, const char *url, int type, unsigned int flags,
 				__LINE__, gnutls_strerror(ret));
 			exit(1);
 		}
-		fprintf(outfile, "\tID: %s\n\n", buf);
+		fprintf(outfile, "\tID: %s\n", buf);
+
+		if (otype == GNUTLS_PKCS11_OBJ_X509_CRT && print_exts > 0) {
+			ret = gnutls_pkcs11_obj_get_exts(crt_list[i], &exts, &exts_size, 0);
+			if (ret >= 0 && exts_size > 0) {
+				gnutls_datum_t txt;
+
+				if (print_exts > 1) {
+					fprintf(outfile, "\tAttached extensions:\n");
+					ret = gnutls_x509_ext_print(exts, exts_size, 0, &txt);
+					if (ret >= 0) {
+						fprintf(outfile, "%s", (char*)txt.data);
+						gnutls_free(txt.data);
+					}
+				} else {
+					fprintf(outfile, "\tAttached extensions:");
+					for (j=0;j<exts_size;j++) {
+						fprintf(outfile, "%s%s", exts[j].oid, (j!=exts_size-1)?",":" ");
+					}
+				}
+				for (j=0;j<exts_size;j++) {
+					gnutls_x509_ext_deinit(&exts[j]);
+				}
+				gnutls_free(exts);
+				fprintf(outfile, "\n");
+			}
+		}
+
+		fprintf(outfile, "\n");
 	}
 
 	return;
 }
 
+#define TEST_DATA "Test data to sign"
+
+void
+pkcs11_test_sign(FILE * outfile, const char *url, unsigned int flags,
+	    common_info_st * info)
+{
+	gnutls_privkey_t privkey;
+	gnutls_pubkey_t pubkey;
+	int ret;
+	gnutls_datum_t data, sig = {NULL, 0};
+	int pk;
+
+	pkcs11_common(info);
+
+	FIX(url, outfile, 0, info);
+
+	data.data = (void*)TEST_DATA;
+	data.size = sizeof(TEST_DATA)-1;
+
+	ret = gnutls_privkey_init(&privkey);
+	if (ret < 0) {
+		fprintf(stderr, "Error in %s:%d: %s\n", __func__,
+			__LINE__, gnutls_strerror(ret));
+		exit(1);
+	}
+
+	ret = gnutls_pubkey_init(&pubkey);
+	if (ret < 0) {
+		fprintf(stderr, "Error in %s:%d: %s\n", __func__,
+			__LINE__, gnutls_strerror(ret));
+		exit(1);
+	}
+
+	ret = gnutls_privkey_import_url(privkey, url, flags);
+	if (ret < 0) {
+		fprintf(stderr, "Cannot import private key: %s\n",
+			gnutls_strerror(ret));
+		exit(1);
+	}
+
+	ret = gnutls_pubkey_import_privkey(pubkey, privkey, GNUTLS_KEY_DIGITAL_SIGNATURE, flags);
+	if (ret < 0) {
+		fprintf(stderr, "Cannot import public key: %s\n",
+			gnutls_strerror(ret));
+		exit(1);
+	}
+
+	ret = gnutls_privkey_sign_data(privkey, GNUTLS_DIG_SHA1, 0, &data, &sig);
+	if (ret < 0) {
+		fprintf(stderr, "Cannot sign data: %s\n",
+			gnutls_strerror(ret));
+		exit(1);
+	}
+
+	pk = gnutls_pubkey_get_pk_algorithm(pubkey, NULL);
+
+	fprintf(stderr, "Verifying against private key parameters... ");
+	ret = gnutls_pubkey_verify_data2(pubkey, gnutls_pk_to_sign(pk, GNUTLS_DIG_SHA1),
+		0, &data, &sig);
+	if (ret < 0) {
+		fprintf(stderr, "Cannot verify signed data: %s\n",
+			gnutls_strerror(ret));
+		exit(1);
+	}
+
+	fprintf(stderr, "ok\n");
+
+	/* now try to verify against a public key within the token */
+	gnutls_pubkey_deinit(pubkey);
+	ret = gnutls_pubkey_init(&pubkey);
+	if (ret < 0) {
+		fprintf(stderr, "Error in %s:%d: %s\n", __func__,
+			__LINE__, gnutls_strerror(ret));
+		exit(1);
+	}
+
+	ret = gnutls_pubkey_import_url(pubkey, url, flags);
+	if (ret < 0) {
+		fprintf(stderr, "Cannot find a corresponding public key object in token: %s\n",
+			gnutls_strerror(ret));
+		if (ret == GNUTLS_E_REQUESTED_DATA_NOT_AVAILABLE)
+			exit(0);
+		exit(1);
+	}
+
+	fprintf(stderr, "Verifying against public key in the token... ");
+	ret = gnutls_pubkey_verify_data2(pubkey, gnutls_pk_to_sign(pk, GNUTLS_DIG_SHA1),
+		0, &data, &sig);
+	if (ret < 0) {
+		fprintf(stderr, "Cannot verify signed data: %s\n",
+			gnutls_strerror(ret));
+		exit(1);
+	}
+
+	fprintf(stderr, "ok\n");
+
+	gnutls_free(sig.data);
+	gnutls_pubkey_deinit(pubkey);
+	gnutls_privkey_deinit(privkey);
+}
+
 void
 pkcs11_export(FILE * outfile, const char *url, unsigned int flags,
 	      common_info_st * info)
@@ -219,7 +359,7 @@ pkcs11_export(FILE * outfile, const char *url, unsigned int flags,
 			gnutls_strerror(ret));
 		exit(1);
 	}
-	
+
 	ret = gnutls_pkcs11_obj_export3(obj, info->outcert_format, &t);
 	if (ret < 0) {
         	fprintf(stderr, "Error in %s:%d: %s\n", __func__,
@@ -273,7 +413,7 @@ pkcs11_export_chain(FILE * outfile, const char *url, unsigned int flags,
 			gnutls_strerror(ret));
 		exit(1);
 	}
-	
+
 	ret = gnutls_x509_crt_import_pkcs11(xcrt, obj);
 	if (ret < 0) {
         	fprintf(stderr, "Error in %s:%d: %s\n", __func__,
@@ -305,9 +445,9 @@ pkcs11_export_chain(FILE * outfile, const char *url, unsigned int flags,
                 
         	fwrite(t.data, 1, t.size, outfile);
                	fputs("\n\n", outfile);
-               	
+
                	gnutls_x509_crt_deinit(xcrt);
-               	
+
                	ret = gnutls_x509_crt_init(&xcrt);
         	if (ret < 0) {
                 	fprintf(stderr, "Error in %s:%d: %s\n", __func__,
@@ -359,6 +499,29 @@ char *get_single_token_url(common_info_st * info)
 	return url;
 }
 
+static
+void print_type(FILE *outfile, unsigned flags)
+{
+	unsigned print = 0;
+
+	fputs("\tType: ", outfile);
+	if (flags & GNUTLS_PKCS11_TOKEN_HW) {
+		fputs("Hardware token", outfile);
+		print++;
+	}
+
+	if (flags & GNUTLS_PKCS11_TOKEN_TRUSTED) {
+		if (print != 0)
+			fputs(", ", outfile);
+		fputs("Trust module", outfile);
+		print++;
+	}
+
+	if (print == 0)
+		fputs("Generic token", outfile);
+	fputc('\n', outfile);
+}
+
 void
 pkcs11_token_list(FILE * outfile, unsigned int detailed,
 		  common_info_st * info, unsigned brief)
@@ -368,6 +531,7 @@ pkcs11_token_list(FILE * outfile, unsigned int detailed,
 	char *url;
 	char buf[128];
 	size_t size;
+	unsigned flags;
 
 	pkcs11_common(info);
 
@@ -383,7 +547,7 @@ pkcs11_token_list(FILE * outfile, unsigned int detailed,
 		}
 
 		if (brief != 0) {
-			fprintf(outfile, "Token %d: %s\n", i, url);
+			fprintf(outfile, "%s\n", url);
 			goto cont;
 		} else {
 			fprintf(outfile, "Token %d:\n\tURL: %s\n", i, url);
@@ -402,6 +566,14 @@ pkcs11_token_list(FILE * outfile, unsigned int detailed,
 
 		fprintf(outfile, "\tLabel: %s\n", buf);
 
+		ret = gnutls_pkcs11_token_get_flags(url, &flags);
+		if (ret < 0) {
+			fprintf(stderr, "Error in %s:%d: %s\n", __func__,
+				__LINE__, gnutls_strerror(ret));
+		} else {
+			print_type(outfile, flags);
+		}
+
 		size = sizeof(buf);
 		ret =
 		    gnutls_pkcs11_token_get_info(url,
@@ -440,6 +612,15 @@ pkcs11_token_list(FILE * outfile, unsigned int detailed,
 		}
 
 		fprintf(outfile, "\tSerial: %s\n", buf);
+
+		size = sizeof(buf);
+		ret =
+		    gnutls_pkcs11_token_get_info(url,
+						 GNUTLS_PKCS11_TOKEN_MODNAME,
+						 buf, &size);
+		if (ret >= 0) {
+			fprintf(outfile, "\tModule: %s\n", buf);
+		}
 		fprintf(outfile, "\n\n");
  cont:
 		gnutls_free(url);
@@ -451,13 +632,16 @@ pkcs11_token_list(FILE * outfile, unsigned int detailed,
 
 void
 pkcs11_write(FILE * outfile, const char *url, const char *label,
-	     unsigned flags, common_info_st * info)
+	     const char *id, unsigned flags, common_info_st * info)
 {
 	gnutls_x509_crt_t xcrt;
 	gnutls_x509_privkey_t xkey;
 	int ret;
-	unsigned int key_usage = 0;
 	gnutls_datum_t *secret_key;
+	unsigned key_usage = 0;
+	unsigned char raw_id[128];
+	size_t raw_id_size;
+	gnutls_datum_t cid = {NULL, 0};
 
 	pkcs11_common(info);
 
@@ -468,11 +652,22 @@ pkcs11_write(FILE * outfile, const char *url, const char *label,
 		label = read_str("warning: The object's label was not specified.\nLabel: ");
 	}
 
+	if (id != NULL) {
+		raw_id_size = sizeof(raw_id);
+		ret = gnutls_hex2bin(id, strlen(id), raw_id, &raw_id_size);
+		if (ret < 0) {
+			fprintf(stderr, "Error converting hex: %s\n", gnutls_strerror(ret));
+			exit(1);
+		}
+		cid.data = raw_id;
+		cid.size = raw_id_size;
+	}
+
 	secret_key = load_secret_key(0, info);
 	if (secret_key != NULL) {
 		ret =
 		    gnutls_pkcs11_copy_secret_key(url, secret_key, label,
-						  key_usage,
+						  info->key_usage,
 						  flags |
 						  GNUTLS_PKCS11_OBJ_FLAG_MARK_SENSITIVE);
 		if (ret < 0) {
@@ -484,7 +679,7 @@ pkcs11_write(FILE * outfile, const char *url, const char *label,
 
 	xcrt = load_cert(0, info);
 	if (xcrt != NULL) {
-		ret = gnutls_pkcs11_copy_x509_crt(url, xcrt, label, flags);
+		ret = gnutls_pkcs11_copy_x509_crt2(url, xcrt, label, &cid, flags);
 		if (ret < 0) {
 			fprintf(stderr, "Error writing certificate: %s\n", gnutls_strerror(ret));
 			if (((flags & GNUTLS_PKCS11_OBJ_FLAG_MARK_CA) ||
@@ -500,10 +695,10 @@ pkcs11_write(FILE * outfile, const char *url, const char *label,
 	xkey = load_x509_private_key(0, info);
 	if (xkey != NULL) {
 		ret =
-		    gnutls_pkcs11_copy_x509_privkey(url, xkey, label,
-						    key_usage,
-						    flags |
-						    GNUTLS_PKCS11_OBJ_FLAG_MARK_SENSITIVE);
+		    gnutls_pkcs11_copy_x509_privkey2(url, xkey, label,
+						     &cid, key_usage|info->key_usage,
+						     flags |
+						     GNUTLS_PKCS11_OBJ_FLAG_MARK_SENSITIVE);
 		if (ret < 0) {
 			fprintf(stderr, "Error in %s:%d: %s\n", __func__,
 				__LINE__, gnutls_strerror(ret));
@@ -523,21 +718,33 @@ pkcs11_write(FILE * outfile, const char *url, const char *label,
 void
 pkcs11_generate(FILE * outfile, const char *url, gnutls_pk_algorithm_t pk,
 		unsigned int bits,
-		const char *label, int detailed,
+		const char *label, const char *id, int detailed,
 		unsigned int flags, common_info_st * info)
 {
 	int ret;
 	gnutls_datum_t pubkey;
+	gnutls_datum_t cid = {NULL, 0};
+	unsigned char raw_id[128];
+	size_t raw_id_size;
 
 	pkcs11_common(info);
 
 	FIX(url, outfile, detailed, info);
 	CHECK_LOGIN_FLAG(flags);
 
+	if (id != NULL) {
+		raw_id_size = sizeof(raw_id);
+		ret = gnutls_hex2bin(id, strlen(id), raw_id, &raw_id_size);
+		if (ret < 0) {
+			fprintf(stderr, "Error converting hex: %s\n", gnutls_strerror(ret));
+			exit(1);
+		}
+		cid.data = raw_id;
+		cid.size = raw_id_size;
+	}
+
 	if (outfile == stderr || outfile == stdout) {
 		fprintf(stderr, "warning: no --outfile was specified and the generated public key will be printed on screen.\n");
-		fprintf(stderr, "note: in some tokens it is impossible to obtain the public key in any other way after generation.\n");
-		sleep(3);
 	}
 
 	if (label == NULL && info->batch == 0) {
@@ -545,8 +752,9 @@ pkcs11_generate(FILE * outfile, const char *url, gnutls_pk_algorithm_t pk,
 	}
 
 	ret =
-	    gnutls_pkcs11_privkey_generate2(url, pk, bits, label,
+	    gnutls_pkcs11_privkey_generate3(url, pk, bits, label, &cid,
 					    GNUTLS_X509_FMT_PEM, &pubkey,
+					    info->key_usage,
 					    flags);
 	if (ret < 0) {
 		fprintf(stderr, "Error in %s:%d: %s\n", __func__, __LINE__,
@@ -947,3 +1155,56 @@ pkcs11_get_random(FILE * outfile, const char *url, unsigned bytes,
 
 	return;
 }
+
+static
+void pkcs11_set_val(FILE * outfile, const char *url, int detailed,
+		   unsigned int flags, common_info_st * info,
+		   gnutls_pkcs11_obj_info_t val_type, const char *val)
+{
+	int ret;
+	gnutls_pkcs11_obj_t obj;
+
+	pkcs11_common(info);
+
+	FIX(url, outfile, detailed, info);
+	CHECK_LOGIN_FLAG(flags);
+
+	ret = gnutls_pkcs11_obj_init(&obj);
+	if (ret < 0) {
+		fprintf(stderr, "Error in %s:%d: %s\n", __func__, __LINE__,
+			gnutls_strerror(ret));
+		exit(1);
+	}
+
+	ret = gnutls_pkcs11_obj_import_url(obj, url, flags);
+	if (ret < 0) {
+		fprintf(stderr, "Error in %s:%d: %s\n", __func__, __LINE__,
+			gnutls_strerror(ret));
+		exit(1);
+	}
+
+	ret =
+	    gnutls_pkcs11_obj_set_info(obj, val_type, val, strlen(val), flags);
+	if (ret < 0) {
+		fprintf(stderr, "Error in %s:%d: %s\n", __func__, __LINE__,
+			gnutls_strerror(ret));
+		exit(1);
+	}
+	gnutls_pkcs11_obj_deinit(obj);
+
+	return;
+}
+
+void pkcs11_set_id(FILE * outfile, const char *url, int detailed,
+		   unsigned int flags, common_info_st * info,
+		   const char *id)
+{
+	return pkcs11_set_val(outfile, url, detailed, flags, info, GNUTLS_PKCS11_OBJ_ID_HEX, id);
+}
+
+void pkcs11_set_label(FILE * outfile, const char *url, int detailed,
+		   unsigned int flags, common_info_st * info,
+		   const char *label)
+{
+	return pkcs11_set_val(outfile, url, detailed, flags, info, GNUTLS_PKCS11_OBJ_LABEL, label);
+}
